/*
 * 16-segment font editor
 * Matt Sarnoff (www.msarnoff.org)
 * June 2, 2009
 * 
 * Requires SDL and SDL_gfx to build
 *
 * Controls:
 * - toggle segments on/off by clicking them
 * - cycle through characters with up/down arrows or the mouse wheel
 * - Enter saves the pattern and advances to the next character
 * - S saves the pattern without advancing to the next character
 * - R reverts to the previously saved pattern
 * - F1 saves the font as a C array in a file called segments.h
 */ 


#include <stdlib.h>
#include <stdbool.h>

#include <SDL/SDL.h>
#include <SDL/SDL_gfxPrimitives.h>

#define WINDOW_WIDTH  500
#define WINDOW_HEIGHT 300 

#define FONT_WIDTH  10
#define FONT_HEIGHT 20
#define FONT_SIZE   10240
#define FONT_NAME   "10x20.fnt"

#define COLOR_ARGB(r,g,b) ((r<<16) + (g<<8) + (b) + 0xFF000000)
#define BACKGROUND        COLOR_ARGB(0,0,0)
#define MIN(a,b) ((a)>(b)?(b):(a))

#define MAX(a,b) ((a)<(b)?(b):(a))

SDL_Surface *screen;

typedef struct
{
  int sx1, sy1, sx2, sy2;
  int rx1, ry1, rx2, ry2;
} Segment;

static Segment segments[16] = {
  {60,20,120,20,   60,15,120,25},
  {120,20,180,20,  120,15,180,25},
  {180,20,160,120, 175,20,165,120},
  {160,120,140,220,155,120,145,220},
  {140,220,80,220, 140,215,80,225},
  {80,220,20,220,  80,215,20,225},
  {20,220,40,120,  25,220,35,120},
  {40,120,60,20,   45,120,55,20},

  {40,120,100,120, 40,115,100,125},
  {100,120,160,120,100,115,160,125},
  {60,20,100,120,  60,30,100,110},
  {100,120,120,20, 105,120,115,20},
  {100,120,180,20, 120,110,160,30},
  {100,120,140,220,100,130,140,210},
  {100,120,80,220, 95,120,85,220},
  {100,120,20,220, 80,130,40,210}
};

int main(int argc, char **argv)
{
  bool doRead = false;
  char filename[128];
  if (argc == 2)
  {
    strncpy(filename, argv[1], 128);
    doRead = true;
  }
  else
    strcpy(filename, "segments.h");

  if (SDL_Init(SDL_INIT_VIDEO))
    exit(-2);
  if ((screen = SDL_SetVideoMode(WINDOW_WIDTH,WINDOW_HEIGHT,32,
          SDL_HWSURFACE | SDL_DOUBLEBUF)) == NULL)
    exit(-3);

  SDL_WM_SetCaption("Segment Editor", "Segment Editor");

  FILE *fontfile = fopen(FONT_NAME, "r");
  unsigned char *font = (unsigned char *)malloc(FONT_SIZE);
  fread(font, FONT_SIZE, 1, fontfile);
  fclose(fontfile);
  gfxPrimitivesSetFont(font, FONT_WIDTH, FONT_HEIGHT);

  Uint8 *keystate;
  bool run = true;
  Uint16 pattern = 0x0000;

  char currentChar = 0;
  Uint16 charPatterns[96];
  memset(charPatterns, 0, sizeof(charPatterns));

  if (doRead)
  {
    FILE *infile = fopen(filename, "r");
    int i = 0;
    while (!feof(infile) && i <= 95)
    {
      char line[80];
      Uint16 p;
      fgets(line, 79, infile);
      p = strtol(line, NULL, 16);
      charPatterns[i] = p;
      i++;
    }
    fclose(infile);
    pattern = charPatterns[currentChar];
  }

  while (run)
  {
    SDL_Event event;
    while (SDL_PollEvent(&event))
    {
      switch (event.type)
      {
        case SDL_QUIT:
          run = false;
          break;
        case SDL_MOUSEBUTTONDOWN:
        {
          if (event.button.button == SDL_BUTTON_LEFT)
          {
            int i;
            for (i = 0; i < 16; i++)
            {
              Segment *s = segments+i;
              int x = event.button.x;
              int y = event.button.y;
              int xmin = MIN(s->rx1, s->rx2);
              int xmax = MAX(s->rx1, s->rx2);
              int ymin = MIN(s->ry1, s->ry2);
              int ymax = MAX(s->ry1, s->ry2);
              if ((x >= xmin && x <= xmax) &&
                  (y >= ymin && y <= ymax))
              {
                pattern ^= (1 << i);
                break;
              }
            }
          }
          else if (event.button.button == SDL_BUTTON_WHEELUP)
          {
            currentChar++;
            if (currentChar > 95)
              currentChar = 0;
            pattern = charPatterns[currentChar];
          }
          else if (event.button.button == SDL_BUTTON_WHEELDOWN)
          {
            currentChar--;
            if (currentChar < 0)
              currentChar = 95;
            pattern = charPatterns[currentChar];
          }
        }
        case SDL_KEYDOWN:
        {
          switch (event.key.keysym.sym)
          {
            case SDLK_UP:
              currentChar++;
              if (currentChar > 95)
                currentChar = 0;
              pattern = charPatterns[currentChar];
              break;
            case SDLK_DOWN:
              currentChar--;
              if (currentChar < 0)
                currentChar = 95;
              pattern = charPatterns[currentChar];
              break;
            case SDLK_RETURN: // save and advance
              charPatterns[currentChar] = pattern;
              currentChar++;
              if (currentChar > 95)
                currentChar = 0;
              pattern = charPatterns[currentChar];
              break;
            case SDLK_s: // save
              charPatterns[currentChar] = pattern;
              break;
            case SDLK_r: // revert
              pattern = charPatterns[currentChar];
              break;
            case SDLK_x:
              pattern = 0;
              break;
            case SDLK_F1: // write to file
            {
              FILE *outfile = fopen(filename, "w");
              int i = 0;
              for (i = 0; i <= 95; i++)
              {
                fprintf(outfile, "0x%04x,  /* %c */\n", charPatterns[i], ' '+i);
              }
              printf("Wrote to %s\n", filename);
              fclose(outfile);
            }
          }
        }
        break;
      }
    }

    keystate = SDL_GetKeyState(NULL);
    if (keystate[SDLK_ESCAPE])
      run = false;

    if (SDL_MUSTLOCK(screen))
      SDL_LockSurface(screen);

    SDL_FillRect(screen, NULL, BACKGROUND);

    int i;
    for (i = 0; i < 16; i++)
    {
      Segment *s = segments+i;
      Uint32 color = (pattern & (1 << i)) ? 0xFF0000FF : 0x400000FF;
      lineColor(screen,s->sx1, s->sy1, s->sx2, s->sy2, color);
    }

    char hexString[7];
    char binString[17], binString2[17];
    snprintf(hexString, 7, "0x%04x", pattern);
    for (i = 0; i < 16; i++)
      binString[15-i] = (pattern & (1 << i)) ? '1' : '0';
    binString[16] = '\0'; 

    stringColor(screen, 200, 100, hexString, 0xA0A0A0FF);
    stringColor(screen, 200, 120, binString, 0xA0A0A0FF);

    char charString[16];
    snprintf(charString, 16, "%c  %d", currentChar+' ', currentChar+' ');

    for (i = 0; i < 16; i++)
      binString2[15-i] = (charPatterns[currentChar] & (1 << i)) ? '1' : '0';
    binString2[16] = '\0'; 

    stringColor(screen, 200, 180, charString, 0x00A0FFFF);
    stringColor(screen, 200, 200, binString2, 0x00A0FFFF);

    stringColor(screen, 5, 255,
        "s=store  enter=store/advance  r=revert  F1=write",
        0x00A000FF);
    stringColor(screen, 5, 275,
        "segment editor by matt sarnoff - june 2 2009 v1.0",
        0x008000FF);
    
    if (SDL_MUSTLOCK(screen))
      SDL_UnlockSurface(screen);

    SDL_Flip(screen);
  }

  return 0;
}
